#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ncurses.h>
#include <errno.h>
#include <sys/select.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <ctype.h>

extern "C" {
#include "error.h"
}
#include "utils.h"
#include "br.h"
#include "pl.h"
extern int sort_order;
extern int group_by_state;

/*
service {
        host_name=ap
        service_description=HTTP
        modified_attributes=0
        check_command=check_tcp!80
        event_handler=
        has_been_checked=1
        should_be_scheduled=1
        check_execution_time=0.009
        check_latency=0.712
        current_state=0
        last_hard_state=0
        current_attempt=1
        max_attempts=3
        state_type=1
        last_state_change=1128634047
        last_hard_state_change=1128633828
        last_time_ok=1129101510
        last_time_warning=0
        last_time_unknown=1128633828
        last_time_critical=1126954624
        plugin_output=TCP OK - 0.001 second response time on port 80
        performance_data=time=0.001121s;0.000000;0.000000;0.000000;10.000000
        last_check=1129101510
        next_check=1129101630
        check_type=0
        current_notification_number=0
        last_notification=0
        next_notification=0
        no_more_notifications=0
        notifications_enabled=1
        active_checks_enabled=1
        passive_checks_enabled=1
        event_handler_enabled=1
        problem_has_been_acknowledged=0
        acknowledgement_type=0
        flap_detection_enabled=1
        failure_prediction_enabled=1
        process_performance_data=1
        obsess_over_service=1
        last_update=1129101556
        is_flapping=0
        percent_state_change=0.00
        scheduled_downtime_depth=0
        }

host {
        host_name=ap
        modified_attributes=0
        check_command=check-host-alive
        event_handler=
        has_been_checked=1
        should_be_scheduled=0
        check_execution_time=0.012
        check_latency=0.000
        current_state=0
        last_hard_state=0
        check_type=0
        plugin_output=PING OK - Packet loss = 0%, RTA = 1.54 ms
        performance_data=
        last_check=1128863442
        next_check=0
        current_attempt=1
        max_attempts=10
        state_type=1
        last_state_change=1128634057
        last_hard_state_change=1128634057
        last_time_up=1128863442
        last_time_down=1128633868
        last_time_unreachable=0
        last_notification=1128634057
        next_notification=0
        no_more_notifications=0
        current_notification_number=0
        notifications_enabled=1
        problem_has_been_acknowledged=0
        acknowledgement_type=0
        active_checks_enabled=1
        passive_checks_enabled=1
        event_handler_enabled=1
        flap_detection_enabled=1
        failure_prediction_enabled=1
        process_performance_data=1
        obsess_over_host=1
        last_update=1129101556
        is_flapping=0
        percent_state_change=0.00
        scheduled_downtime_depth=0
        }
*/

void parse_2_0_statuslog(int fd, struct stats **pstats, int *n_stats)
{
	int type = TYPE_IGNORE;
	char *host_name = NULL;
	int current_state = -1;
	char *service_description = NULL;
	char *plugin_output = NULL;
	time_t last_state_change = 0;
	char active_checks_enabled = 0;
	char passive_checks_enabled = 0;
	char notifications_enabled = 0;
	char problem_has_been_acknowledged = 0;
	int scheduled_downtime_depth = 0;
	char state_type = 0;
	char last_hard_state = 0;
	int modified_attributes = 0;
	int event_handler = 0;
	int has_been_checked = 0;
	int should_be_scheduled = 0;
	int current_attempt = 0;
	int max_attempts = 0;
	int last_hard_state_change = 0;
	int last_time_ok = 0;
	int last_time_warning = 0;
	int last_time_unknown = 0;
	int last_time_critical = 0;
	int last_check = 0;
	int next_check = 0;
	int check_type = 0;
	int current_notification_number = 0;
	int last_notification = 0;
	int next_notification = 0;
	int no_more_notifications = 0;
	int event_handler_enabled = 0;
	int acknowledgement_type = 0;
	int flap_detection_enabled = 0;
	int failure_prediction_enabled = 0;
	int process_performance_data = 0;
	int obsess_over_service = 0;
	int last_update = 0;
	int is_flapping = 0;
	double percent_state_change = 0;
	double check_execution_time = 0;
	double check_latency = 0;
	char * performance_data = NULL;
	char * check_command = NULL;
	int last_time_up = 0;
	int last_time_down = 0;
	int last_time_unreachable = 0;

	*pstats = NULL;
	*n_stats = -1;

	/* start up buffered reader */
	buffered_reader bf(fd);

	*pstats = NULL;
	*n_stats = 0;

	for(;;)
	{
		char *cb;
		char *line = bf.read_line();
		if (!line)
			break;

		cb = strchr(line, '{');
		if (cb)
		{
			/* init */
			type = TYPE_IGNORE;
			host_name = service_description = plugin_output = performance_data = check_command = NULL;
			percent_state_change = check_execution_time = check_latency = 0;
			last_state_change = 0;
			current_state = -1;
			active_checks_enabled = passive_checks_enabled = notifications_enabled = problem_has_been_acknowledged = state_type = last_hard_state = 0;
			scheduled_downtime_depth = modified_attributes = event_handler = has_been_checked = should_be_scheduled = current_attempt = max_attempts = 0;
			last_hard_state_change = last_time_ok = last_time_warning = last_time_unknown = last_time_critical = last_check = next_check = 0;
			check_type = current_notification_number = last_notification = next_notification = no_more_notifications = event_handler_enabled = 0;
			acknowledgement_type = flap_detection_enabled = failure_prediction_enabled = process_performance_data = obsess_over_service = last_update = 0;
			is_flapping = last_time_up = last_time_down = last_time_unreachable = 0;

			if (strncmp(line, "host ", 5) == 0)
			{
				type = TYPE_HOST;
			}
			else if (strncmp(line, "service", 7) == 0)
			{
				type = TYPE_SERVICE;
			}
			else
			{
				type = TYPE_IGNORE;
			}
		}
		else if (type != TYPE_IGNORE)
		{
			char *is = strchr(line, '=');
			if (is == NULL && strchr(line, '}') != NULL)	/* end of definition, store in array */
			{
				*pstats = (struct stats *)myrealloc(*pstats, ((*n_stats) + 1) * sizeof(struct stats), "stats array");
				if (!*pstats)
					error_exit("Error allocating memory");

				(*pstats)[*n_stats].type = type;
				(*pstats)[*n_stats].host_name = host_name?host_name:mystrdup("");
				if (type == TYPE_HOST && current_state == 1) /* HOSTS: 0=ok, 1=down, 2=unreachable */
					current_state = 2;
				(*pstats)[*n_stats].current_state = current_state;
				(*pstats)[*n_stats].service_description = service_description?service_description:mystrdup("");
				(*pstats)[*n_stats].plugin_output = plugin_output?plugin_output:mystrdup("");
				(*pstats)[*n_stats].last_state_change = last_state_change;
				(*pstats)[*n_stats].active_checks_enabled = active_checks_enabled;
				(*pstats)[*n_stats].passive_checks_enabled = passive_checks_enabled;
				(*pstats)[*n_stats].notifications_enabled = notifications_enabled;
				(*pstats)[*n_stats].problem_has_been_acknowledged = problem_has_been_acknowledged;
				(*pstats)[*n_stats].scheduled_downtime_depth = scheduled_downtime_depth;
				(*pstats)[*n_stats].state_type = state_type;
				(*pstats)[*n_stats].last_hard_state = last_hard_state;
				(*pstats)[*n_stats].modified_attributes = modified_attributes;
				(*pstats)[*n_stats].event_handler = event_handler;
				(*pstats)[*n_stats].has_been_checked = has_been_checked;
				(*pstats)[*n_stats].should_be_scheduled = should_be_scheduled;
				(*pstats)[*n_stats].current_attempt = current_attempt;
				(*pstats)[*n_stats].max_attempts = max_attempts;
				(*pstats)[*n_stats].last_hard_state_change = last_hard_state_change;
				(*pstats)[*n_stats].last_time_ok = last_time_ok;
				(*pstats)[*n_stats].last_time_warning = last_time_warning;
				(*pstats)[*n_stats].last_time_unknown = last_time_unknown;
				(*pstats)[*n_stats].last_time_critical = last_time_critical;
				(*pstats)[*n_stats].last_check = last_check;
				(*pstats)[*n_stats].next_check = next_check;
				(*pstats)[*n_stats].check_type = check_type;
				(*pstats)[*n_stats].current_notification_number = current_notification_number;
				(*pstats)[*n_stats].last_notification = last_notification;
				(*pstats)[*n_stats].next_notification = next_notification;
				(*pstats)[*n_stats].no_more_notifications = no_more_notifications;
				(*pstats)[*n_stats].event_handler_enabled = event_handler_enabled;
				(*pstats)[*n_stats].acknowledgement_type = acknowledgement_type;
				(*pstats)[*n_stats].flap_detection_enabled = flap_detection_enabled;
				(*pstats)[*n_stats].failure_prediction_enabled = failure_prediction_enabled;
				(*pstats)[*n_stats].process_performance_data = process_performance_data;
				(*pstats)[*n_stats].obsess_over_service = obsess_over_service;
				(*pstats)[*n_stats].last_update = last_update;
				(*pstats)[*n_stats].is_flapping = is_flapping;
				(*pstats)[*n_stats].percent_state_change = percent_state_change;
				(*pstats)[*n_stats].check_execution_time = check_execution_time;
				(*pstats)[*n_stats].check_latency = check_latency;
				(*pstats)[*n_stats].performance_data = performance_data;
				(*pstats)[*n_stats].check_command = check_command;
				(*pstats)[*n_stats].last_time_up = last_time_up;
				(*pstats)[*n_stats].last_time_down = last_time_down;
				(*pstats)[*n_stats].last_time_unreachable = last_time_unreachable;
				(*n_stats)++;

				check_command = host_name = performance_data = service_description = plugin_output = NULL;
			}
			else if (is != NULL)
			{
				char *cmd = line, *par = is + 1;
				while(*cmd == ' ' || *cmd == '\t') cmd++;

				*is = 0x00;

				if (strcmp(cmd, "host_name") == 0)
					host_name = mystrdup(par);
				else if (strcmp(cmd, "current_state") == 0)
					current_state = atoi(par);
				else if (strcmp(cmd, "service_description") == 0)
					service_description = mystrdup(par);
				else if (strcmp(cmd, "plugin_output") == 0)
					plugin_output = mystrdup(par);
				else if (strcmp(cmd, "last_state_change") == 0)
					last_state_change = atol(par);
				else if (strcmp(cmd, "active_checks_enabled") == 0)
					active_checks_enabled = atoi(par);
				else if (strcmp(cmd, "passive_checks_enabled") == 0)
					passive_checks_enabled = atoi(par);
				else if (strcmp(cmd, "notifications_enabled") == 0)
					notifications_enabled = atoi(par);
				else if (strcmp(cmd, "problem_has_been_acknowledged") == 0)
					problem_has_been_acknowledged = atoi(par);
				else if (strcmp(cmd, "scheduled_downtime_depth") == 0)
					scheduled_downtime_depth = atoi(par);
				else if (strcmp(cmd, "state_type") == 0)
					state_type = atoi(par);
				else if (strcmp(cmd, "last_hard_state") == 0)
					last_hard_state = atoi(par);
				else if (strcmp(cmd, "modified_attributes") == 0)
					modified_attributes = atoi(par);
				else if (strcmp(cmd, "event_handler") == 0)
					event_handler = atoi(par);
				else if (strcmp(cmd, "has_been_checked") == 0)
					has_been_checked = atoi(par);
				else if (strcmp(cmd, "should_be_scheduled") == 0)
					should_be_scheduled = atoi(par);
				else if (strcmp(cmd, "current_attempt") == 0)
					current_attempt = atoi(par);
				else if (strcmp(cmd, "max_attempts") == 0)
					max_attempts = atoi(par);
				else if (strcmp(cmd, "last_hard_state_change") == 0)
					last_hard_state_change = atoi(par);
				else if (strcmp(cmd, "last_time_ok") == 0)
					last_time_ok = atoi(par);
				else if (strcmp(cmd, "last_time_warning") == 0)
					last_time_warning = atoi(par);
				else if (strcmp(cmd, "last_time_unknown") == 0)
					last_time_unknown = atoi(par);
				else if (strcmp(cmd, "last_time_critical") == 0)
					last_time_critical = atoi(par);
				else if (strcmp(cmd, "last_check") == 0)
					last_check = atoi(par);
				else if (strcmp(cmd, "next_check") == 0)
					next_check = atoi(par);
				else if (strcmp(cmd, "check_type") == 0)
					check_type = atoi(par);
				else if (strcmp(cmd, "current_notification_number") == 0)
					current_notification_number = atoi(par);
				else if (strcmp(cmd, "last_notification") == 0)
					last_notification = atoi(par);
				else if (strcmp(cmd, "next_notification") == 0)
					next_notification = atoi(par);
				else if (strcmp(cmd, "no_more_notifications") == 0)
					no_more_notifications = atoi(par);
				else if (strcmp(cmd, "event_handler_enabled") == 0)
					event_handler_enabled = atoi(par);
				else if (strcmp(cmd, "acknowledgement_type") == 0)
					acknowledgement_type = atoi(par);
				else if (strcmp(cmd, "flap_detection_enabled") == 0)
					flap_detection_enabled = atoi(par);
				else if (strcmp(cmd, "failure_prediction_enabled") == 0)
					failure_prediction_enabled = atoi(par);
				else if (strcmp(cmd, "process_performance_data") == 0)
					process_performance_data = atoi(par);
				else if (strcmp(cmd, "obsess_over_service") == 0)
					obsess_over_service = atoi(par);
				else if (strcmp(cmd, "last_update") == 0)
					last_update = atoi(par);
				else if (strcmp(cmd, "is_flapping") == 0)
					is_flapping = atoi(par);
				else if (strcmp(cmd, "percent_state_change") == 0)
					percent_state_change = atof(par);
				else if (strcmp(cmd, "check_execution_time") == 0)
					check_execution_time = atof(par);
				else if (strcmp(cmd, "check_latency") == 0)
					check_latency = atof(par);
				else if (strcmp(cmd, "performance_data") == 0)
					performance_data = mystrdup(par);
				else if (strcmp(cmd, "check_command") == 0)
					check_command = mystrdup(par);
				else if (strcmp(cmd, "last_time_up") == 0)
					last_time_up = atoi(par);
				else if (strcmp(cmd, "last_time_down") == 0)
					last_time_down = atoi(par);
				else if (strcmp(cmd, "last_time_unreachable") == 0)
					last_time_unreachable = atoi(par);
			}
		}

		/* this one does not use myfree() as the buffered reader does not
		 * use mymalloc() and friends
		 */
		free(line);
	}
}

int split_1_0_line(char *line, char **out)
{
	int out_index = 0;

	memset(out, 0x00, sizeof(char *) * V1_0_MAX_ELEMENTS);

	for(;out_index < V1_0_MAX_ELEMENTS;)
	{
		out[out_index++] = line;

		while(*line != ';' && *line != 0x00)
			line++;

		if (*line == 0x00)
			break;

		*line = 0x00; /* replace ';' with 0x00 */
		line++;
	}

	return out_index;
}

void parse_1_0_statuslog(int fd, struct stats **pstats, int *n_stats)
{
	/* start up buffered reader */
	buffered_reader bf(fd);

	*pstats = NULL;
	*n_stats = 0;

	for(;;)
	{
		int type = TYPE_IGNORE;
		char *host_name = NULL;
		int current_state = -1;
		char *service_description = NULL;
		char *plugin_output = NULL;
		time_t last_state_change = 0;
		char active_checks_enabled = 0;
		char passive_checks_enabled = 0;
		char notifications_enabled = 0;
		char problem_has_been_acknowledged = 0;
		int scheduled_downtime_depth = 0;
		char state_type = 0;
		char last_hard_state = 0;
		int modified_attributes = 0;
		int event_handler = 0;
		int has_been_checked = 0;
		int should_be_scheduled = 0;
		int current_attempt = 0;
		int max_attempts = 0;
		int last_hard_state_change = 0;
		int last_time_ok = 0;
		int last_time_warning = 0;
		int last_time_unknown = 0;
		int last_time_critical = 0;
		int last_check = 0;
		int next_check = 0;
		int check_type = 0;
		int current_notification_number = 0;
		int last_notification = 0;
		int next_notification = 0;
		int no_more_notifications = 0;
		int event_handler_enabled = 0;
		int acknowledgement_type = 0;
		int flap_detection_enabled = 0;
		int failure_prediction_enabled = 0;
		int process_performance_data = 0;
		int obsess_over_service = 0;
		int last_update = 0;
		int is_flapping = 0;
		double percent_state_change = 0;
		double check_execution_time = 0;
		double check_latency = 0;
		char * performance_data = NULL;
		char * check_command = NULL;
		int last_time_up = 0;
		int last_time_down = 0;
		int last_time_unreachable = 0;

		char *dummy;
		char *elements[V1_0_MAX_ELEMENTS];
		int n_elem;
		char *line = bf.read_line();
		if (!line)
			break;

		if (line[0] == '#')
			goto skip_line;

		n_elem = split_1_0_line(line, elements);

		*pstats = (struct stats *)myrealloc(*pstats, ((*n_stats) + 1) * sizeof(struct stats), "stats array");
		if (!*pstats)
			error_exit("Error allocating memory");

/*
[Time of last update] HOST;
Host Name (string);
Status (OK/DOWN/UNREACHABLE);
Last Check Time (long time);
Last State Change (long time);
Acknowledged (0/1);
Time Up (long time);
Time Down (long time);
Time Unreachable (long time);
Last Notification Time (long time);
Current Notification Number (#);
Notifications Enabled (0/1);
Event Handlers Enabled (0/1);
Checks Enabled (0/1);
Flap Detection Enabled (0/1);
Host is Flapping (0/1);
Percent State Change (###.##);
Scheduled downtime depth (#);
Failure Prediction Enabled (0/1);
Process Performance Data(0/1);
Plugin Output (string)
*/

/*
[Time of last update] SERVICE;
Host Name (string);
Service Description (string);
Status (OK/WARNING/CRITICAL/UNKNOWN);
Retry number (#/#);
State Type (SOFT/HARD);
Last check time (long time);
Next check time (long time);
Check type (ACTIVE/PASSIVE);
Checks enabled (0/1);
Accept Passive Checks (0/1);
Event Handlers Enabled (0/1);
Last state change (long time);
Problem acknowledged (0/1);
Last Hard State (OK/WARNING/CRITICAL/UNKNOWN);
Time OK (long time);
Time Unknown (long time);
Time Warning (long time);
Time Critical (long time);
Last Notification Time (long time);
Current Notification Number (#);
Notifications Enabled (0/1);
Latency (#);
Execution Time (#);
Flap Detection Enabled (0/1);
Service is Flapping (0/1);
Percent State Change (###.##);
Scheduled Downtime Depth (#);
Failure Prediction Enabled (0/1);
Process Performance Date (0/1);
Obsess Over Service (0/1);
Plugin Output (string)
*/
	
		dummy = strchr(elements[0], ' ');
		if (!dummy)
			goto skip_line;
		dummy++;
		if (strcmp(dummy, "HOST") == 0)
			type = TYPE_HOST;
		else if (strcmp(dummy, "SERVICE") == 0)
			type = TYPE_SERVICE;
		else if (strcmp(dummy, "PROGRAM") == 0)
			goto skip_line;
		else
			goto skip_line;

		if (type == TYPE_HOST)
		{
			if (n_elem != 21)
				goto skip_line;

			host_name = mystrdup(elements[1]);
			if (strcmp(elements[2], "OK"))
				current_state = 0;
			else
				current_state = 2;
			last_check = atol(elements[3]);
			last_state_change = atol(elements[4]);
			problem_has_been_acknowledged = atoi(elements[5]);
			last_time_up = atol(elements[6]);
			last_time_down = atol(elements[7]);
			last_time_unreachable = atol(elements[8]);
			last_notification = atol(elements[9]);
			current_notification_number = atoi(elements[10]);
			notifications_enabled = atoi(elements[11]);
			event_handler_enabled = atoi(elements[12]);
			active_checks_enabled = atoi(elements[13]); /* in 2.0 it has been split up in passive and active */
			flap_detection_enabled = atoi(elements[14]);
			is_flapping = atoi(elements[15]);
			percent_state_change = atof(elements[16]);
			scheduled_downtime_depth = atoi(elements[17]);
			failure_prediction_enabled = atoi(elements[18]);
/*			process_performance_data = atoi(elements[19]); */
			plugin_output = mystrdup(elements[20]);
		}
		else if (type == TYPE_SERVICE)
		{
			if (n_elem != 32)
				goto skip_line;
			host_name = mystrdup(elements[1]);
			service_description = mystrdup(elements[2]);
			if (strcmp(elements[3], "OK") == 0)
				current_state = 0;
			else if (strcmp(elements[3], "WARNING") == 0)
				current_state = 1;
			else if (strcmp(elements[3], "CRITICAL") == 0)
				current_state = 2;
			else if (strcmp(elements[3], "UNKNOWN") == 0)
				current_state = 3;
/*			retry_number = atoi(elements[4]); */
			if (strcmp(elements[5], "SOFT") == 0)
				state_type = 0;
			else
				state_type = 1;
			last_check = atol(elements[6]);
			next_check = atol(elements[7]);
			if (strcmp(elements[8], "ACTIVE") == 0)
				check_type = 1;
			else
				check_type = 0;
			active_checks_enabled = atoi(elements[9]);
/*			accept_passive_checks = atoi(elements[10]); */
			event_handler_enabled = atoi(elements[11]);
			last_state_change = atol(elements[12]);
			problem_has_been_acknowledged = atoi(elements[13]);
			if (strcmp(elements[14], "OK") == 0)
				last_hard_state = 0;
			else if (strcmp(elements[14], "WARNING") == 0)
				last_hard_state = 1;
			else if (strcmp(elements[14], "CRITICAL") == 0)
				last_hard_state = 2;
			else if (strcmp(elements[14], "UNKNOWN") == 0)
				last_hard_state = 3;
			last_time_ok = atol(elements[15]);
			last_time_unknown = atol(elements[16]);
			last_time_warning = atol(elements[17]);
			last_time_critical = atol(elements[18]);
			last_notification = atol(elements[19]);
			current_notification_number = atoi(elements[20]);
			notifications_enabled = atoi(elements[21]);
			check_latency = atof(elements[22]);
			check_execution_time = atof(elements[23]);
			flap_detection_enabled = atoi(elements[24]);
			is_flapping = atoi(elements[25]);
			percent_state_change = atof(elements[26]);
			scheduled_downtime_depth = atoi(elements[27]);
			failure_prediction_enabled = atoi(elements[28]);
/*			process_performance_date = atoi(elements[29]); */
			obsess_over_service = atoi(elements[30]);
			plugin_output = mystrdup(elements[31]);
		}
		else
			error_exit("internal error: type %d unexpected\n", type);

		(*pstats)[*n_stats].type = type;
		(*pstats)[*n_stats].host_name = host_name?host_name:mystrdup("");
		if (type == TYPE_HOST && current_state == 1) /* HOSTS: 0=ok, 1=down, 2=unreachable */
			current_state = 2;
		(*pstats)[*n_stats].current_state = current_state;
		(*pstats)[*n_stats].service_description = service_description?service_description:mystrdup("");
		(*pstats)[*n_stats].plugin_output = plugin_output?plugin_output:mystrdup("");
		(*pstats)[*n_stats].last_state_change = last_state_change;
		(*pstats)[*n_stats].active_checks_enabled = active_checks_enabled;
		(*pstats)[*n_stats].passive_checks_enabled = passive_checks_enabled;
		(*pstats)[*n_stats].notifications_enabled = notifications_enabled;
		(*pstats)[*n_stats].problem_has_been_acknowledged = problem_has_been_acknowledged;
		(*pstats)[*n_stats].scheduled_downtime_depth = scheduled_downtime_depth;
		(*pstats)[*n_stats].state_type = state_type;
		(*pstats)[*n_stats].last_hard_state = last_hard_state;
		(*pstats)[*n_stats].modified_attributes = modified_attributes;
		(*pstats)[*n_stats].event_handler = event_handler;
		(*pstats)[*n_stats].has_been_checked = has_been_checked;
		(*pstats)[*n_stats].should_be_scheduled = should_be_scheduled;
		(*pstats)[*n_stats].current_attempt = current_attempt;
		(*pstats)[*n_stats].max_attempts = max_attempts;
		(*pstats)[*n_stats].last_hard_state_change = last_hard_state_change;
		(*pstats)[*n_stats].last_time_ok = last_time_ok;
		(*pstats)[*n_stats].last_time_warning = last_time_warning;
		(*pstats)[*n_stats].last_time_unknown = last_time_unknown;
		(*pstats)[*n_stats].last_time_critical = last_time_critical;
		(*pstats)[*n_stats].last_check = last_check;
		(*pstats)[*n_stats].next_check = next_check;
		(*pstats)[*n_stats].check_type = check_type;
		(*pstats)[*n_stats].current_notification_number = current_notification_number;
		(*pstats)[*n_stats].last_notification = last_notification;
		(*pstats)[*n_stats].next_notification = next_notification;
		(*pstats)[*n_stats].no_more_notifications = no_more_notifications;
		(*pstats)[*n_stats].event_handler_enabled = event_handler_enabled;
		(*pstats)[*n_stats].acknowledgement_type = acknowledgement_type;
		(*pstats)[*n_stats].flap_detection_enabled = flap_detection_enabled;
		(*pstats)[*n_stats].failure_prediction_enabled = failure_prediction_enabled;
		(*pstats)[*n_stats].process_performance_data = process_performance_data;
		(*pstats)[*n_stats].obsess_over_service = obsess_over_service;
		(*pstats)[*n_stats].last_update = last_update;
		(*pstats)[*n_stats].is_flapping = is_flapping;
		(*pstats)[*n_stats].percent_state_change = percent_state_change;
		(*pstats)[*n_stats].check_execution_time = check_execution_time;
		(*pstats)[*n_stats].check_latency = check_latency;
		(*pstats)[*n_stats].performance_data = performance_data;
		(*pstats)[*n_stats].check_command = check_command;
		(*pstats)[*n_stats].last_time_up = last_time_up;
		(*pstats)[*n_stats].last_time_down = last_time_down;
		(*pstats)[*n_stats].last_time_unreachable = last_time_unreachable;
		(*n_stats)++;

skip_line:
		free(line);
	}

	close(fd);
}

void free_stats_array(struct stats *pstats, int n_stats)
{
	if (pstats)
	{
		for(int loop=0; loop<n_stats; loop++)
		{
			myfree(pstats[loop].host_name, "entry hostname");
			myfree(pstats[loop].service_description, "entry servicedescription");
			myfree(pstats[loop].plugin_output, "entry plugin output");
			myfree(pstats[loop].performance_data, "entry performance data");
			myfree(pstats[loop].check_command, "check command");
		}

		myfree(pstats, "stats array");
	}
}

int sort_compare_func(const void *a, const void *b)
{
	struct stats *p1 = (struct stats *)a;
	struct stats *p2 = (struct stats *)b;
	int result;

	if(group_by_state && p1 -> current_state != p2 -> current_state)
		return p2 -> current_state - p1 -> current_state;
	switch(sort_order)
	{
		/* Sort by last state change, name, service */
		case 0:
			if (p1 -> last_state_change != p2 -> last_state_change)
				return p2 -> last_state_change - p1 -> last_state_change;
			if((result = strcmp(p1 -> host_name, p2 -> host_name)) != 0)
				return result;
			if((result = strcmp(p1 -> service_description, p2 -> service_description)) != 0)
				return result;
			break;

		/* Sort by name, service, last state change */
		case 1:
			if((result = strcmp(p1 -> host_name, p2 -> host_name)) != 0)
				return result;
			if((result = strcmp(p1 -> service_description, p2 -> service_description)) != 0)
				return result;
			if (p1 -> last_state_change != p2 -> last_state_change)
				return p2 -> last_state_change - p1 -> last_state_change;
			break;
		/* Sort by service, host, last state change */
		case 2:
			if((result = strcmp(p1 -> service_description, p2 -> service_description)) != 0)
				return result;
			if((result = strcmp(p1 -> host_name, p2 -> host_name)) != 0)
				return result;
			if (p1 -> last_state_change != p2 -> last_state_change)
				return p2 -> last_state_change - p1 -> last_state_change;
			break;

	}

	return 0;
}

void sort_stats_array(struct stats *pstats, int n_stats)
{
	if (n_stats > 1)
		qsort(pstats, n_stats, sizeof(struct stats), sort_compare_func);
}

int host_is_down(struct stats *pstats, int n_stats, char *host_name)
{
	for(int loop=0; loop<n_stats; loop++)
	{
		if (pstats[loop].type == TYPE_SERVICE)
			continue;

		if (pstats[loop].current_state == 1 || // down
				pstats[loop].current_state == 2) // unreachable
		{
			if (strcmp(pstats[loop].host_name, host_name) == 0)
				return loop;
		}
	}

	return -1;
}

int should_i_show_entry(struct stats *pstats, int n_stats, int cur_index, char list_all_problems, char always_notify, char also_acknowledged, char hide_ok)
{
	if (pstats[cur_index].state_type != 1)
		return 0;

	if (!list_all_problems)
	{
		if (host_is_down(pstats, n_stats, pstats[cur_index].host_name) != -1 && pstats[cur_index].type != TYPE_HOST)
			return 0;

		if (pstats[cur_index].active_checks_enabled == 0 &&
				pstats[cur_index].passive_checks_enabled == 0)
			return 0;

		if (pstats[cur_index].scheduled_downtime_depth != 0)
			return 0;
	}

	if (!always_notify && pstats[cur_index].notifications_enabled == 0)
		return 0;

	if (!also_acknowledged && pstats[cur_index].problem_has_been_acknowledged == 1)
		return 0;

	if (pstats[cur_index].current_state == 0 && hide_ok)
		return 0;

	if (pstats[cur_index].current_state < 0 || pstats[cur_index].current_state > 3)
		error_exit("internal error: state %d not known", pstats[cur_index].current_state);

	return 1;
}

int find_index_by_host_and_service(struct stats *pstats, int n_stats, char *host_name, char *service_description)
{
	int loop;

	for(loop=0; loop<n_stats; loop++)
	{
		if (strcmp(pstats[loop].host_name, host_name) == 0 &&
		    strcmp(pstats[loop].service_description, service_description) == 0)
		{
			return loop;
		}
	}

	return -1;
}
